from __future__ import division, print_function, absolute_import

import numpy
import pandas
from numpy.random.mtrand import RandomState
from sklearn.metrics.pairwise import pairwise_distances
from hep_ml import commonutils
from hep_ml.commonutils import weighted_quantile, build_normalizer, \
    compute_cut_for_efficiency, generate_sample, compute_knn_indices_of_signal, \
    compute_knn_indices_of_same_class


def test_splitting(n_rows=10, n_columns=8):
    column_names = ['col' + str(i) for i in range(n_columns)]
    signal_df = pandas.DataFrame(numpy.ones([n_rows, n_columns]), columns=column_names)
    bg_df = pandas.DataFrame(numpy.zeros([n_rows, n_columns]), columns=column_names)

    common_X = pandas.concat([signal_df, bg_df], ignore_index=True)
    common_y = numpy.concatenate([numpy.ones(len(signal_df)), numpy.zeros(len(bg_df))])

    trainX, testX, trainY, testY = commonutils.train_test_split(common_X, common_y)

    for (index, row), label in zip(trainX.iterrows(), trainY):
        assert numpy.all(row == label), 'wrong data partition'
    for (index, row), label in zip(testX.iterrows(), testY):
        assert numpy.all(row == label), 'wrong data partition'

    assert (trainX.columns == column_names).all(), 'new column names!'
    assert (testX.columns == column_names).all(), 'new column names!'
    assert len(trainX) + len(testX) == len(common_X), 'new size is strange'


def check_weighted_percentile(size=100, q_size=20):
    random = RandomState()
    array = random.permutation(size)
    quantiles = random.uniform(size=q_size)
    q_permutation = random.permutation(q_size)
    result1 = weighted_quantile(array, quantiles)[q_permutation]
    result2 = weighted_quantile(array, quantiles[q_permutation])
    result3 = weighted_quantile(array[random.permutation(size)], quantiles[q_permutation])
    assert numpy.all(result1 == result2) and numpy.all(result1 == result3), 'breaks on permutations'

    # checks that order is kept
    quantiles = numpy.linspace(0, 1, size * 3)
    x = weighted_quantile(array, quantiles, sample_weight=random.exponential(size=size))
    assert numpy.all(x == numpy.sort(x)), "doesn't preserve order"

    array = numpy.array([0, 1, 2, 5])
    # comparing with simple percentiles
    for x in random.uniform(size=10):
        assert numpy.abs(numpy.percentile(array, x * 100) - weighted_quantile(array, x, old_style=True)) < 1e-7, \
            "doesn't coincide with numpy.percentile"


def test_weighted_percentile():
    check_weighted_percentile(RandomState().randint(4, 40), RandomState().randint(4, 40))
    check_weighted_percentile(100, 20)
    check_weighted_percentile(20, 100)


def test_build_normalizer(checks=10):
    predictions = numpy.array(RandomState().normal(size=2000))
    result = build_normalizer(predictions)(predictions)
    assert numpy.all(result[numpy.argsort(predictions)] == sorted(result))
    assert numpy.all(result >= 0) and numpy.all(result <= 1)
    percentiles = [100 * (i + 1.) / (checks + 1.) for i in range(checks)]
    assert numpy.all(abs(numpy.percentile(result, percentiles) - numpy.array(percentiles) / 100.) < 0.01)

    # testing with weights
    predictions = numpy.exp(predictions / 2)
    weighted_normalizer = build_normalizer(predictions, sample_weight=predictions)
    result = weighted_normalizer(predictions)
    assert numpy.all(result[numpy.argsort(predictions)] == sorted(result))
    assert numpy.all(result >= 0) and numpy.all(result <= 1 + 1e-7)
    predictions = numpy.sort(predictions)
    result = weighted_normalizer(predictions)
    result2 = numpy.cumsum(predictions) / numpy.sum(predictions)
    assert numpy.all(numpy.abs(result - result2) < 0.005)


def test_compute_cut():
    random = RandomState()
    predictions = random.permutation(100)
    labels = numpy.ones(100)
    for eff in [0.1, 0.5, 0.75, 0.99]:
        cut = compute_cut_for_efficiency(eff, labels, predictions)
        assert numpy.sum(predictions > cut) / len(predictions) == eff, 'the cut was set wrongly'

    weights = numpy.array(random.exponential(size=100))
    for eff in random.uniform(size=100):
        cut = compute_cut_for_efficiency(eff, labels, predictions, sample_weight=weights)
        lower = numpy.sum(weights[predictions > cut + 1]) / numpy.sum(weights)
        upper = numpy.sum(weights[predictions > cut - 1]) / numpy.sum(weights)
        assert lower < eff < upper, 'the cut was set wrongly'


def test_compute_knn_indices(n_events=100):
    X, y = generate_sample(n_events, 10, distance=.5)
    is_signal = y > 0.5
    signal_indices = numpy.where(is_signal)[0]
    uniform_columns = X.columns[:1]
    knn_indices = compute_knn_indices_of_signal(X[uniform_columns], is_signal, 10)
    distances = pairwise_distances(X[uniform_columns])
    for i, neighbours in enumerate(knn_indices):
        assert numpy.all(is_signal[neighbours]), "returned indices are not signal"
        not_neighbours = [x for x in signal_indices if not x in neighbours]
        min_dist = numpy.min(distances[i, not_neighbours])
        max_dist = numpy.max(distances[i, neighbours])
        assert min_dist >= max_dist, "distances are set wrongly!"

    knn_all_indices = compute_knn_indices_of_same_class(X[uniform_columns], is_signal, 10)
    for i, neighbours in enumerate(knn_all_indices):
        assert numpy.all(is_signal[neighbours] == is_signal[i]), "returned indices are not signal/bg"




from __future__ import print_function, division, absolute_import
from collections import OrderedDict

import numpy as np
from six.moves import zip
from sklearn.metrics import roc_auc_score, accuracy_score
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
# sklearn 0.22 deprecated sklearn.ensemble.weight_boosting
# remain backwards compatible
try:
    from sklearn.ensemble import AdaBoostClassifier
except ImportError as e:
    from sklearn.ensemble.weight_boosting import AdaBoostClassifier

from hep_ml.commonutils import generate_sample
from hep_ml.metrics import BinBasedCvM, KnnBasedCvM
from hep_ml.uboost import uBoostBDT, uBoostClassifier


def test_cuts(n_samples=1000):
    base_classifier = DecisionTreeClassifier(min_samples_leaf=10, max_depth=6)
    trainX, trainY = generate_sample(n_samples, 10, 0.6)
    uniform_features = ['column0']

    for algorithm in ['SAMME', 'SAMME.R']:
        for target_efficiency in [0.1, 0.3, 0.5, 0.7, 0.9]:
            uBDT = uBoostBDT(
                uniform_features=uniform_features,
                uniform_label=1,
                target_efficiency=target_efficiency,
                n_neighbors=20, n_estimators=20,
                algorithm=algorithm,
                base_estimator=base_classifier)
            uBDT.fit(trainX, trainY)

            passed = sum(trainY) * target_efficiency

            assert uBDT.score_cut == uBDT.score_cuts_[-1], \
                'something wrong with computed cuts'

            for score, cut in zip(uBDT.staged_decision_function(trainX[trainY > 0.5]),
                                  uBDT.score_cuts_):
                passed_upper = np.sum(score > cut - 1e-7)
                passed_lower = np.sum(score > cut + 1e-7)
                assert passed_lower <= passed <= passed_upper, "wrong stage cuts"


def test_probas(n_samples=1000):
    trainX, trainY = generate_sample(n_samples, 10, 0.6)
    testX, testY = generate_sample(n_samples, 10, 0.6)

    params = {
        'n_neighbors': 10,
        'n_estimators': 10,
        'uniform_features': ['column0'],
        'uniform_label': 1,
        'base_estimator': DecisionTreeClassifier(max_depth=5)
    }

    for algorithm in ['SAMME', 'SAMME.R']:
        uboost_classifier = uBoostClassifier(
            algorithm=algorithm,
            efficiency_steps=3, **params)

        bdt_classifier = uBoostBDT(algorithm=algorithm, **params)

        for classifier in [bdt_classifier, uboost_classifier]:
            classifier.fit(trainX, trainY)
            proba1 = classifier.predict_proba(testX)
            proba2 = list(classifier.staged_predict_proba(testX))[-1]
            assert np.allclose(proba1, proba2, atol=0.001), \
                "staged_predict doesn't coincide with the predict for proba."

        score1 = bdt_classifier.decision_function(testX)
        score2 = list(bdt_classifier.staged_decision_function(testX))[-1]
        assert np.allclose(score1, score2), \
            "staged_score doesn't coincide with the score."

        assert len(bdt_classifier.feature_importances_) == trainX.shape[1]


def test_quality(n_samples=3000):
    testX, testY = generate_sample(n_samples, 10, 0.6)
    trainX, trainY = generate_sample(n_samples, 10, 0.6)

    params = {
        'n_neighbors': 10,
        'n_estimators': 10,
        'uniform_features': ['column0'],
        'uniform_label': 1,
        'base_estimator': DecisionTreeClassifier(min_samples_leaf=20, max_depth=5)
    }

    for algorithm in ['SAMME', 'SAMME.R']:
        uboost_classifier = uBoostClassifier(
            algorithm=algorithm, efficiency_steps=5, **params)

        bdt_classifier = uBoostBDT(algorithm=algorithm, **params)

        for classifier in [bdt_classifier, uboost_classifier]:
            classifier.fit(trainX, trainY)
            predict_proba = classifier.predict_proba(testX)
            predict = classifier.predict(testX)
            assert roc_auc_score(testY, predict_proba[:, 1]) > 0.7, \
                "quality is awful"
            print("Accuracy = %.3f" % accuracy_score(testY, predict))


def check_classifiers(n_samples=10000):
    """
    This function is not tested by default, it should be called manually
    """
    testX, testY = generate_sample(n_samples, 10, 0.6)
    trainX, trainY = generate_sample(n_samples, 10, 0.6)
    uniform_features = ['column0']

    ada = AdaBoostClassifier(n_estimators=50)
    ideal_bayes = GaussianNB()

    uBoost_SAMME = uBoostClassifier(
        uniform_features=uniform_features,
        uniform_label=1,
        n_neighbors=50,
        efficiency_steps=5,
        n_estimators=50,
        algorithm="SAMME")

    uBoost_SAMME_R = uBoostClassifier(
        uniform_features=uniform_features,
        uniform_label=1,
        n_neighbors=50,
        efficiency_steps=5,
        n_estimators=50,
        algorithm="SAMME.R")

    uBoost_SAMME_R_threaded = uBoostClassifier(
        uniform_features=uniform_features,
        uniform_label=1,
        n_neighbors=50,
        efficiency_steps=5,
        n_estimators=50,
        n_threads=3,
        subsample=0.9,
        algorithm="SAMME.R")

    clf_dict = OrderedDict({
        "Ada": ada,
        "uBOOST": uBoost_SAMME,
        "uBOOST.R": uBoost_SAMME_R,
        "uBOOST.R2": uBoost_SAMME_R_threaded
    })

    cvms = {}
    for clf_name, clf in clf_dict.items():
        clf.fit(trainX, trainY)
        p = clf.predict_proba(testX)
        metric = KnnBasedCvM(uniform_features=uniform_features)
        metric.fit(testX, testY)
        cvms[clf_name] = metric(testY, p, sample_weight=np.ones(len(testY)))

    assert cvms['uBOOST'] < cvms['ada']
    print(cvms)



from __future__ import division, print_function, absolute_import
import numpy
from hep_ml import splot

__author__ = 'Alex Rogozhnikov'


def test_splot(n_samples=100):
    """
    Checking basic properties of sWeights based on small samples
    """
    for weights in [None, numpy.ones(n_samples), numpy.random.exponential(size=n_samples)]:
        # None is the same as sample_weight = 1
        real_weights = weights
        if real_weights is None:
            real_weights = numpy.ones(n_samples)

        for n_classes in [2, 3, 4, 5]:
            p = numpy.random.random([n_samples, n_classes])
            p /= numpy.sum(p, axis=1, keepdims=True)

            sWeights = splot.compute_sweights(probabilities=p, sample_weight=weights)
            initial_stats = numpy.sum(p * real_weights[:, numpy.newaxis], axis=0)

            # How much inside each bin
            prob_collection = numpy.dot(sWeights.T, p)
            assert numpy.allclose(prob_collection, numpy.diag(initial_stats)), \
                'wrong reconstruction after reweighting (biased)'
            assert numpy.allclose(sWeights.sum(axis=0), initial_stats), \
                'sum of sWeights should be equal to weighted amount of expected samples'
            assert numpy.allclose(sWeights.sum(axis=1), real_weights), \
                'sum of sWeights should be equal to initial weight of sample'



"""
tests for hep_ml.speedup module
"""
from __future__ import division, print_function, absolute_import

import numpy
import pandas
import time
from hep_ml.speedup import LookupClassifier
from hep_ml.commonutils import generate_sample
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.metrics import roc_auc_score
from collections import OrderedDict
from nose.tools import raises

__author__ = 'Alex Rogozhnikov'


def test_lookup(n_samples=10000, n_features=7, n_bins=8):
    X, y = generate_sample(n_samples=n_samples, n_features=n_features, distance=0.6)

    base_estimator = GradientBoostingClassifier()
    clf = LookupClassifier(base_estimator=base_estimator, n_bins=n_bins, keep_trained_estimator=True).fit(X, y)
    p = clf.predict_proba(X)
    assert roc_auc_score(y, p[:, 1]) > 0.8, 'quality of classification is too low'
    assert p.shape == (n_samples, 2)
    assert numpy.allclose(p.sum(axis=1), 1), 'probabilities are not summed up to 1'

    # checking conversions
    lookup_size = n_bins ** n_features
    lookup_indices = numpy.arange(lookup_size, dtype=int)
    bins_indices = clf.convert_lookup_index_to_bins(lookup_indices=lookup_indices)
    lookup_indices2 = clf.convert_bins_to_lookup_index(bins_indices=bins_indices)
    assert numpy.allclose(lookup_indices, lookup_indices2), 'something wrong with conversions'
    assert len(clf._lookup_table) == n_bins ** n_features, 'wrong size of lookup table'

    # checking speed
    X = pandas.concat([X] * 10)
    start = time.time()
    p1 = clf.trained_estimator.predict_proba(clf.transform(X))
    time_old = time.time() - start
    start = time.time()
    p2 = clf.predict_proba(X)
    time_new = time.time() - start
    print(time_old, ' now takes ', time_new)
    assert numpy.allclose(p1, p2), "pipeline doesn't work as expected"


def test_sizes(n_samples=10000, n_features=4, n_bins=8):
    X, y = generate_sample(n_samples=n_samples, n_features=n_features, distance=0.6)

    base_estimator = GradientBoostingClassifier(n_estimators=1)
    clf = LookupClassifier(base_estimator=base_estimator, n_bins=n_bins).fit(X, y)

    bin_indices = clf.transform(X)
    assert numpy.allclose(numpy.max(bin_indices, axis=0) + 1, n_bins)

    maximals = OrderedDict()
    for column in X.columns:
        maximals[column] = numpy.random.randint(low=n_bins // 2, high=n_bins)

    clf = LookupClassifier(base_estimator=base_estimator, n_bins=maximals).fit(X, y)
    bin_indices = clf.transform(X)
    assert numpy.allclose(numpy.max(bin_indices, axis=0) + 1, list(maximals.values()))

    assert numpy.allclose(numpy.min(bin_indices, axis=0), 0)


@raises(ValueError)
def test_raising_exception():
    X, y = generate_sample(n_samples=100, n_features=10)
    LookupClassifier(GradientBoostingClassifier(), n_bins=16).fit(X, y)


def test_classifier_with_dataframe():
    try:
        from rep.estimators import SklearnClassifier
        clf = SklearnClassifier(GradientBoostingClassifier(n_estimators=1))
        X, y = generate_sample(n_samples=100, n_features=4)
        for X_ in [X, pandas.DataFrame(X)]:
            lookup = LookupClassifier(clf, n_bins=16).fit(X_, y)
            lookup.predict_proba(X)
    except ImportError:
        print('expected fail: yandex/rep not installed')



from __future__ import division, print_function, absolute_import
import numpy
from hep_ml.preprocessing import IronTransformer, BinTransformer
from hep_ml.commonutils import generate_sample

__author__ = 'Alex Rogozhnikov'


def assure_monotonic(original, transformed):
    """
    Checks that transformation of data was monotonic
    """
    original = numpy.array(original)
    transformed = numpy.array(transformed)

    assert numpy.shape(original) == numpy.shape(transformed)
    for column in range(original.shape[1]):
        sorter = numpy.argsort(original[:, column])
        sorted_trans = transformed[sorter, column]
        assert numpy.all(numpy.diff(sorted_trans) >= 0)


def test_bin_transform(n_features=10, n_samples=10000):
    """
    Testing BinTransformer
    """
    data = numpy.random.normal(size=[n_samples, n_features])

    n_bins = 41

    transformer = BinTransformer(max_bins=n_bins).fit(data)
    result = transformer.transform(data)

    assert numpy.all(result < n_bins)
    assert numpy.all(result >= 0)
    assert numpy.allclose(numpy.max(result, axis=0), n_bins - 1)
    assert numpy.allclose(numpy.min(result, axis=0), 0)
    assure_monotonic(data, result)

    # check reproducibility
    assert numpy.all(transformer.transform(data) == transformer.transform(data))

    # checking dtype is integer
    numpy_result = numpy.array(result)
    print(numpy_result.dtype)
    assert numpy_result.dtype == 'uint8'


def test_iron_transformer(n_features=10, n_samples=5000):
    """
    Testing IronTransformer
    """
    import pandas
    data1 = numpy.random.normal(size=[n_samples, n_features])
    data2 = pandas.DataFrame(data=numpy.random.normal(size=[n_samples, n_features]),
                             index=numpy.random.choice(n_samples * 2, size=n_samples, replace=False))
    data3 = numpy.clip(data1, -0.2, 0.2)

    for symmetrize in [True, False]:
        for max_points in [n_samples // 2, n_samples * 2]:
            for data in [data1, data2, data3]:
                data_copy = data.copy()

                transformer = IronTransformer(max_points=max_points, symmetrize=symmetrize).fit(data)
                result = transformer.transform(data)

                assert numpy.all(data == data_copy), 'data was augmented!'

                assert numpy.all(numpy.isfinite(result))  # , numpy.isfinite(result).all()
                assert numpy.all(result <= 1.)
                if symmetrize:
                    assert numpy.all(result >= -1)
                else:
                    assert numpy.all(result >= 0.)
                assure_monotonic(data, result)

                # check reproducibility
                assert numpy.all(transformer.transform(data) == transformer.transform(data))

                # checking dtype is integer
                numpy_result = numpy.array(result)
                assert numpy_result.dtype == float
                for name, (feature_values, feature_percentiles) in transformer.feature_maps.items():
                    assert len(feature_values) <= max_points


def test_bin_transformer_limits(n_features=10, n_bins=123):
    X, y = generate_sample(n_samples=1999, n_features=n_features)
    X = BinTransformer(max_bins=n_bins).fit_transform(X)
    assert numpy.allclose(X.max(axis=0), n_bins - 1)

    X_orig, y = generate_sample(n_samples=20, n_features=n_features)
    X = BinTransformer(max_bins=n_bins).fit_transform(X_orig)
    assert numpy.allclose(X.min(axis=0), 0)


def test_bin_transformer_extend_to(n_features=10, n_bins=123):
    extended_length = 19
    X, y = generate_sample(n_samples=20, n_features=n_features)
    X1 = BinTransformer(max_bins=n_bins).fit(X).transform(X)
    X2 = BinTransformer(max_bins=n_bins).fit(X).transform(X, extend_to=extended_length)
    assert len(X2) % extended_length == 0, 'wrong shape!'
    assert numpy.allclose(X2[:len(X1)], X1), 'extending does not work as expected!'



from __future__ import division, print_function, absolute_import

import numpy
from hep_ml import losses
from hep_ml.commonutils import generate_sample

__author__ = 'Alex Rogozhnikov'


def test_orders(size=40):
    effs1 = losses._compute_positions(numpy.arange(size), numpy.ones(size))
    p = numpy.random.permutation(size)
    effs2 = losses._compute_positions(numpy.arange(size)[p], numpy.ones(size))
    assert numpy.all(effs1[p] == effs2), 'Efficiencies are wrong'
    assert numpy.all(effs1 == numpy.sort(effs1)), 'sortings are wrong'


def test_loss_functions(size=50, epsilon=1e-3):
    """
    Testing that Hessians and gradients of loss functions coincide with numerical approximations
    """
    X, y = generate_sample(size, n_features=10)
    rank_column = X.columns[2]
    X[rank_column] = numpy.random.randint(0, 3, size=size)
    sample_weight = numpy.random.exponential(size=size)
    tested_losses = [
        losses.MSELossFunction(),
        losses.MAELossFunction(),
        losses.LogLossFunction(),
        losses.AdaLossFunction(),
        losses.KnnAdaLossFunction(X.columns[:1], uniform_label=1, knn=5),
        losses.CompositeLossFunction(),
        losses.RankBoostLossFunction(rank_column),
    ]
    pred = numpy.random.normal(size=size)
    # y = pred is a special point in i.e. MAELossFunction
    pred[numpy.abs(y - pred) < epsilon] = - 0.1
    print(sum(numpy.abs(y - pred) < epsilon))

    for loss in tested_losses:
        loss.fit(X, y, sample_weight=sample_weight)
        # testing sign of gradient
        val = loss(pred)
        gradient = loss.negative_gradient(pred)

        numer_gradient = numpy.zeros(len(pred))
        numer_hessian = numpy.zeros(len(pred))
        for i in range(size):
            pred_plus = pred.copy()
            pred_plus[i] += epsilon
            val_plus = loss(pred_plus)

            pred_minus = pred.copy()
            pred_minus[i] -= epsilon
            val_minus = loss(pred_minus)

            numer_gradient[i] = - (val_plus - val_minus) / 2. / epsilon
            numer_hessian[i] = (val_plus + val_minus - 2 * val) / epsilon ** 2

        assert numpy.allclose(gradient, numer_gradient), 'wrong computation of gradient for {}'.format(loss)
        if not isinstance(loss, losses.MSELossFunction) and not isinstance(loss, losses.MAELossFunction):
            assert (gradient * (2 * y - 1) >= 0).all(), 'wrong signs of gradients'
        if isinstance(loss, losses.HessianLossFunction):
            hessian = loss.hessian(pred)
            assert numpy.allclose(hessian, numer_hessian, atol=1e-5), 'wrong computation of hessian for {}'.format(loss)


def test_step_optimality(n_samples=100):
    """
    testing that for single leaf function returns the optimal value
    """
    X, y = generate_sample(n_samples, n_features=10)
    sample_weight = numpy.random.exponential(size=n_samples)

    rank_column = X.columns[2]
    X[rank_column] = numpy.random.randint(0, 3, size=n_samples)

    tested_losses = [
        losses.LogLossFunction(),
        losses.AdaLossFunction(),
        losses.KnnAdaLossFunction(X.columns[:1], uniform_label=0, knn=5),
        losses.CompositeLossFunction(),
        losses.RankBoostLossFunction(rank_column),
        losses.MSELossFunction(),
    ]

    pred = numpy.random.normal(size=n_samples)

    for loss in tested_losses:
        loss.fit(X, y, sample_weight=sample_weight)

        # Test simple way to get optimal step
        leaf_value = numpy.random.normal()
        step = 0.
        for _ in range(4):
            ministep, = loss.prepare_new_leaves_values(terminal_regions=numpy.zeros(n_samples, dtype=int),
                                                       leaf_values=[leaf_value], y_pred=pred + step)
            step += ministep

        if isinstance(loss, losses.MAELossFunction):
            # checking that MAE is minimized with long process
            for iteration in range(1, 30):
                ministep, = loss.prepare_new_leaves_values(terminal_regions=numpy.zeros(n_samples, dtype=int),
                                                           leaf_values=[leaf_value], y_pred=pred + step)
                step += ministep * 1. / iteration

        loss_values = []
        coeffs = [0.9, 1.0, 1.1]
        for coeff in coeffs:
            loss_values.append(loss(pred + coeff * step))
        print(loss, step, 'losses: ', loss_values)
        assert loss_values[1] <= loss_values[0] + 1e-7
        assert loss_values[1] <= loss_values[2] + 1e-7

        # Test standard function
        opt_value = loss.compute_optimal_step(y_pred=pred)
        loss_values2 = []
        for coeff in coeffs:
            loss_values2.append(loss(pred + coeff * opt_value))
        print(loss, step, 'losses: ', loss_values)
        assert loss_values2[1] <= loss_values2[0] + 1e-7
        assert loss_values2[1] <= loss_values2[2] + 1e-7



from __future__ import division, print_function, absolute_import

import numpy

from hep_ml.reweight import BinsReweighter, GBReweighter, FoldingReweighter
from hep_ml.metrics_utils import ks_2samp_weighted

__author__ = 'Alex Rogozhnikov'


def weighted_covariance(data, weights):
    if len(data.shape) == 1:
        data = data[:, numpy.newaxis]
    data = data - numpy.mean(data, axis=0, keepdims=True)
    weights = weights * 1. / weights.sum()
    return numpy.einsum('ij, ik, i -> jk', data, data, weights)


def check_reweighter(n_dimensions, n_samples, reweighter, folding=False):
    mean_original = numpy.random.normal(size=n_dimensions)
    cov_original = numpy.diag([1.] * n_dimensions)

    mean_target = numpy.random.mtrand.multivariate_normal(mean=mean_original, cov=cov_original)
    cov_target = cov_original * 0.4 + numpy.ones([n_dimensions, n_dimensions]) * 0.2

    original = numpy.random.mtrand.multivariate_normal(mean=mean_original, cov=cov_original, size=n_samples + 1)
    original_weight = numpy.ones(n_samples + 1)

    target = numpy.random.mtrand.multivariate_normal(mean=mean_target, cov=cov_target, size=n_samples)
    target_weight = numpy.ones(n_samples)

    reweighter.fit(original, target, original_weight=original_weight, target_weight=target_weight)
    new_weights_array = []
    new_weights_array.append(reweighter.predict_weights(original, original_weight=original_weight))
    if folding:
        def mean_vote(x):
            return numpy.mean(x, axis=0)

        new_weights_array.append(reweighter.predict_weights(original, original_weight=original_weight,
                                                            vote_function=mean_vote))

    for new_weights in new_weights_array:
        av_orig = numpy.average(original, weights=original_weight, axis=0)
        print('WAS', av_orig)
        av_now = numpy.average(original, weights=new_weights, axis=0)
        print('NOW:', av_now)
        av_ideal = numpy.average(target, weights=target_weight, axis=0)
        print('IDEAL:', av_ideal)

        print('COVARIANCE')
        print('WAS', weighted_covariance(original, original_weight))
        print('NOW', weighted_covariance(original, new_weights))
        print('IDEAL', weighted_covariance(target, target_weight))

        assert numpy.all(abs(av_now - av_ideal) < abs(av_orig - av_ideal)), 'averages are wrong'
        for dim in range(n_dimensions):
            diff1 = ks_2samp_weighted(original[:, dim], target[:, dim], original_weight, target_weight)
            diff2 = ks_2samp_weighted(original[:, dim], target[:, dim], new_weights, target_weight)
            print('KS', diff1, diff2)
            assert diff2 < diff1, 'Differences {} {}'.format(diff1, diff2)


def test_reweighter_1d():
    reweighter = BinsReweighter(n_bins=200, n_neighs=2)
    check_reweighter(n_dimensions=1, n_samples=100000, reweighter=reweighter)


def test_gb_reweighter_1d():
    reweighter = GBReweighter(n_estimators=100, max_depth=2)
    check_reweighter(n_dimensions=1, n_samples=100000, reweighter=reweighter)


def test_reweighter_2d():
    reweighter = BinsReweighter(n_bins=20, n_neighs=2)
    check_reweighter(n_dimensions=2, n_samples=1000000, reweighter=reweighter)


def test_gb_reweighter_2d():
    reweighter = GBReweighter(max_depth=3, n_estimators=30, learning_rate=0.3, gb_args=dict(subsample=0.3))
    check_reweighter(n_dimensions=2, n_samples=200000, reweighter=reweighter)


def test_folding_gb_reweighter():
    reweighter = FoldingReweighter(GBReweighter(n_estimators=20, max_depth=2, learning_rate=0.1), n_folds=3)
    check_reweighter(n_dimensions=2, n_samples=200000, reweighter=reweighter, folding=True)


def test_folding_bins_reweighter():
    reweighter = FoldingReweighter(BinsReweighter(n_bins=20, n_neighs=2), n_folds=3)
    check_reweighter(n_dimensions=2, n_samples=1000000, reweighter=reweighter, folding=True)



from __future__ import division, print_function, absolute_import

import numpy
import pandas
from scipy.stats import ks_2samp

from numpy.random.mtrand import RandomState
from hep_ml.commonutils import generate_sample
from hep_ml.metrics_utils import prepare_distribution, _ks_2samp_fast, ks_2samp_weighted, _cvm_2samp_fast, \
    group_indices_to_groups_matrix
from hep_ml.metrics import KnnBasedSDE, KnnBasedTheil, KnnBasedCvM, \
    BinBasedSDE, BinBasedTheil, BinBasedCvM
from hep_ml.metrics_utils import bin_to_group_indices, compute_bin_indices
from tests._metrics_oldimplementation import compute_sde_on_bins, compute_sde_on_groups, compute_theil_on_bins, \
    compute_theil_on_groups, bin_based_ks, groups_based_ks, cvm_2samp, bin_based_cvm, group_based_cvm, sde, \
    cvm_flatness, theil_flatness

__author__ = 'Alex Rogozhnikov'

random = RandomState()


def generate_binned_dataset(n_samples, n_bins):
    """useful function, generates dataset with bins, groups, random weights.
    This is used to test correlation functions. """
    random = RandomState()
    y = random.uniform(size=n_samples) > 0.5
    pred = random.uniform(size=(n_samples, 2))
    weights = random.exponential(size=(n_samples,))
    bins = random.randint(0, n_bins, n_samples)
    groups = bin_to_group_indices(bin_indices=bins, mask=(y == 1))
    return y, pred, weights, bins, groups


def test_bin_to_group_indices(size=100, bins=10):
    bin_indices = RandomState().randint(0, bins, size=size)
    mask = RandomState().randint(0, 2, size=size) > 0.5
    group_indices = bin_to_group_indices(bin_indices, mask=mask)
    assert numpy.sum([len(group) for group in group_indices]) == numpy.sum(mask)
    a = numpy.sort(numpy.concatenate(group_indices))
    b = numpy.where(mask > 0.5)[0]
    assert numpy.all(a == b), 'group indices are computed wrongly'


def test_groups_matrix(size=1000, bins=4):
    bin_indices = RandomState().randint(0, bins, size=size)
    mask = RandomState().randint(0, 2, size=size) > 0.5
    n_signal_events = numpy.sum(mask)
    group_indices = bin_to_group_indices(bin_indices, mask=mask)
    assert numpy.sum([len(group) for group in group_indices]) == n_signal_events
    group_matrix = group_indices_to_groups_matrix(group_indices, n_events=size)
    assert group_matrix.sum() == n_signal_events
    for event_id, (is_signal, bin) in enumerate(zip(mask, bin_indices)):
        assert group_matrix[bin, event_id] == is_signal


def test_bins(size=500, n_bins=10):
    columns = ['var1', 'var2']
    df = pandas.DataFrame(random.uniform(size=(size, 2)), columns=columns)
    x_limits = numpy.linspace(0, 1, n_bins + 1)[1:-1]
    bins = compute_bin_indices(df[columns].values, bin_limits=[x_limits, x_limits])
    assert numpy.all(0 <= bins) and numpy.all(bins < n_bins * n_bins), "the bins with wrong indices appeared"


def test_compare_sde_computations(n_samples=1000, n_bins=10):
    y, pred, weights, bins, groups = generate_binned_dataset(n_samples=n_samples, n_bins=n_bins)
    target_efficiencies = RandomState().uniform(size=3)
    a = compute_sde_on_bins(pred[:, 1], mask=(y == 1), bin_indices=bins,
                            target_efficiencies=target_efficiencies, sample_weight=weights)
    b = compute_sde_on_groups(pred[:, 1], mask=(y == 1), groups_indices=groups,
                              target_efficiencies=target_efficiencies, sample_weight=weights)
    assert numpy.allclose(a, b)


def test_theil(n_samples=1000, n_bins=10):
    y, pred, weights, bins, groups = generate_binned_dataset(n_samples=n_samples, n_bins=n_bins)
    a = compute_theil_on_bins(pred[:, 1], y == 1, bins, [0.5, 0.78], sample_weight=weights)
    b = compute_theil_on_groups(pred[:, 1], y == 1, groups, [0.5, 0.78], sample_weight=weights)
    assert numpy.allclose(a, b)


def test_ks2samp_fast(size=1000):
    y1 = RandomState().uniform(size=size)
    y2 = y1[RandomState().uniform(size=size) > 0.5]
    a = ks_2samp(y1, y2)[0]
    prep_data, prep_weights, prep_F = prepare_distribution(y1, numpy.ones(len(y1)))
    b = _ks_2samp_fast(prep_data, y2, prep_weights, numpy.ones(len(y2)), cdf1=prep_F)
    c = _ks_2samp_fast(prep_data, y2, prep_weights, numpy.ones(len(y2)), cdf1=prep_F)
    d = ks_2samp_weighted(y1, y2, numpy.ones(len(y1)) / 3, numpy.ones(len(y2)) / 4)
    assert numpy.allclose(a, b, rtol=1e-2, atol=1e-3)
    assert numpy.allclose(b, c)
    assert numpy.allclose(b, d)
    print('ks2samp is ok')


def test_ks(n_samples=1000, n_bins=10):
    y, pred, weights, bins, groups = generate_binned_dataset(n_samples=n_samples, n_bins=n_bins)
    mask = y == 1
    a = bin_based_ks(pred[:, 1], mask=mask, sample_weight=weights, bin_indices=bins)
    b = groups_based_ks(pred[:, 1], mask=mask, sample_weight=weights, groups_indices=groups)
    assert numpy.allclose(a, b)


def test_ks2samp(n_samples1=100, n_samples2=100):
    """
    checking that KS can be computed with ROC curve
    """
    data1 = numpy.random.normal(size=n_samples1)
    weights1 = numpy.random.random(size=n_samples1)
    data2 = numpy.random.normal(size=n_samples2)
    weights2 = numpy.random.random(size=n_samples2)

    print(weights1.sum(), 'SUM')

    KS = ks_2samp_weighted(data1, data2, weights1=weights1, weights2=weights2)

    # alternative way to check
    labels = [0] * len(data1) + [1] * len(data2)
    data = numpy.concatenate([data1, data2])
    weights = numpy.concatenate([weights1, weights2])
    from sklearn.metrics import roc_curve
    fpr, tpr, _ = roc_curve(labels, data, sample_weight=weights)
    KS2 = numpy.max(numpy.abs(symmetrize(fpr) - symmetrize(tpr)))
    print(KS, KS2)
    print(weights1.sum(), 'SUM')
    assert numpy.allclose(KS, KS2), 'different values of KS'


def symmetrize(F):
    return 0.5 * (F + numpy.insert(F[:-1], 0, [0]))


def test_cvm2samp(n_samples1=100, n_samples2=100):
    data1 = numpy.random.normal(size=n_samples1)
    weights1 = numpy.random.random(size=n_samples1)
    data2 = numpy.random.normal(size=n_samples2)
    weights2 = numpy.random.random(size=n_samples2)

    CVM = cvm_2samp(data1, data2, weights1=weights1, weights2=weights2)

    # alternative way to check
    labels = [0] * len(data1) + [1] * len(data2)
    data = numpy.concatenate([data1, data2])
    weights = numpy.concatenate([weights1, weights2])
    from sklearn.metrics import roc_curve
    fpr, tpr, _ = roc_curve(labels, data, sample_weight=weights)
    # data1 corresponds to
    weights1 = numpy.diff(numpy.insert(fpr, 0, [0]))
    CVM2 = numpy.sum(weights1 * (symmetrize(fpr) - symmetrize(tpr)) ** 2)
    print(CVM, CVM2)
    assert numpy.allclose(CVM, CVM2), 'different values of CVM'


def test_fast_cvm(n_samples=1000):
    random = RandomState()
    data1 = random.uniform(size=n_samples)
    weights1 = random.uniform(size=n_samples)
    mask = random.uniform(size=n_samples) > 0.5
    data2 = data1[mask]
    weights2 = weights1[mask]
    a = cvm_2samp(data1, data2, weights1, weights2)
    prepared_data1, prepared_weights1, F1 = prepare_distribution(data1, weights1)
    b = _cvm_2samp_fast(prepared_data1, data2, prepared_weights1, weights2, cdf1=F1)
    assert numpy.allclose(a, b)


def test_cvm(size=1000):
    y_pred = random.uniform(size=size)
    y = random.uniform(size=size) > 0.5
    sample_weight = random.exponential(size=size)
    bin_indices = random.randint(0, 10, size=size)
    mask = y == 1
    groups_indices = bin_to_group_indices(bin_indices=bin_indices, mask=mask)
    cvm1 = bin_based_cvm(y_pred[mask], sample_weight=sample_weight[mask], bin_indices=bin_indices[mask])
    cvm2 = group_based_cvm(y_pred, mask=mask, sample_weight=sample_weight, groups_indices=groups_indices)
    assert numpy.allclose(cvm1, cvm2)


def test_cvm_sde_limit(size=2000):
    """ Checks that in the limit CvM coincides with MSE """
    effs = numpy.linspace(0, 1, 2000)
    y_pred = random.uniform(size=size)
    y = random.uniform(size=size) > 0.5
    sample_weight = random.exponential(size=size)
    bin_indices = random.randint(0, 10, size=size)
    y_pred += bin_indices * random.uniform()
    mask = y == 1

    val1 = bin_based_cvm(y_pred[mask], sample_weight=sample_weight[mask], bin_indices=bin_indices[mask])
    val2 = compute_sde_on_bins(y_pred, mask=mask, bin_indices=bin_indices, target_efficiencies=effs,
                               sample_weight=sample_weight)

    assert numpy.allclose(val1, val2 ** 2, atol=1e-3, rtol=1e-2)


def test_new_metrics(n_samples=2000, knn=50):
    X, y = generate_sample(n_samples=n_samples, n_features=10)
    sample_weight = numpy.random.exponential(size=n_samples) ** 0.
    predictions_orig = numpy.random.random(size=[n_samples, 2])

    for shift in [0.1, 0.2]:
        predictions = predictions_orig.copy()
        predictions[:, 1] += shift * y
        predictions /= predictions.sum(axis=1, keepdims=True)

        # Checking SDE
        features = X.columns[:1]
        sde_val1 = sde(y, predictions, X, uniform_features=features, sample_weight=sample_weight, label=0, knn=knn)
        sde_metric = KnnBasedSDE(n_neighbours=knn, uniform_features=features, uniform_label=0, )
        sde_metric.fit(X, y, sample_weight=sample_weight)
        sde_val2 = sde_metric(y, predictions, sample_weight=sample_weight)

        assert numpy.allclose(sde_val1, sde_val2), 'SDE values are different'

        # Checking theil
        theil_val1 = theil_flatness(y, predictions, X, uniform_features=features, sample_weight=sample_weight,
                                    label=0, knn=knn)
        theil_metric = KnnBasedTheil(n_neighbours=knn, uniform_features=features, uniform_label=0, )
        theil_metric.fit(X, y, sample_weight=sample_weight)
        theil_val2 = theil_metric(y, predictions, sample_weight=sample_weight)
        print(theil_val1, theil_val2)
        assert numpy.allclose(theil_val1, theil_val2), 'Theil values are different'

        # Checking CVM
        features = X.columns[:1]
        cvm_val1 = cvm_flatness(y, predictions, X, uniform_features=features, sample_weight=sample_weight, label=0,
                                knn=knn)
        cvm_metric = KnnBasedCvM(n_neighbours=knn, uniform_features=features, uniform_label=0, )
        cvm_metric.fit(X, y, sample_weight=sample_weight)
        cvm_val2 = cvm_metric(y, predictions, sample_weight=sample_weight)

        assert numpy.allclose(cvm_val1, cvm_val2), 'CvM values are different'


def test_metrics_clear(n_samples=2000, knn=50, uniform_class=0):
    """
    Testing that after deleting all inappropriate events (events of other class),
    metrics stays the same
    """
    X, y = generate_sample(n_samples=n_samples, n_features=10)
    sample_weight = numpy.random.exponential(size=n_samples)
    predictions = numpy.random.random(size=[n_samples, 2])
    predictions /= predictions.sum(axis=1, keepdims=True)
    features = X.columns[:1]

    mask = (y == uniform_class)
    X_clear = X[mask]
    y_clear = y[mask]
    sample_weight_clear = sample_weight[mask]
    predictions_clear = predictions[mask]

    for function in [sde, theil_flatness, cvm_flatness]:
        flatness_val = function(y, predictions, X, uniform_features=features, sample_weight=sample_weight, label=0,
                                knn=knn)
        flatness_val_clear = function(y_clear, predictions_clear, X_clear, uniform_features=features,
                                      sample_weight=sample_weight_clear, label=0, knn=knn)
        assert flatness_val == flatness_val_clear, 'after deleting other class, the metrics changed'

    for class_ in [KnnBasedSDE, KnnBasedTheil, KnnBasedCvM]:
        metric1 = class_(n_neighbours=knn, uniform_features=features, uniform_label=0, )
        metric1.fit(X, y, sample_weight=sample_weight)
        flatness_val1 = metric1(y, predictions, sample_weight)

        metric2 = class_(n_neighbours=knn, uniform_features=features, uniform_label=0, )
        metric2.fit(X_clear, y_clear, sample_weight=sample_weight_clear)
        flatness_val2 = metric2(y_clear, predictions_clear, sample_weight_clear)
        assert flatness_val1 == flatness_val2, 'after deleting other class, the metrics changed'


def test_workability(n_samples=2000, knn=50, uniform_label=0, n_bins=10):
    """Simply checks that metrics are working """
    X, y = generate_sample(n_samples=n_samples, n_features=10)
    sample_weight = numpy.random.exponential(size=n_samples)
    predictions = numpy.random.random(size=[n_samples, 2])
    predictions /= predictions.sum(axis=1, keepdims=True)
    features = X.columns[:1]

    for class_ in [KnnBasedSDE, KnnBasedTheil, KnnBasedCvM]:
        metric = class_(n_neighbours=knn, uniform_features=features, uniform_label=uniform_label, )
        metric.fit(X, y, sample_weight=sample_weight)
        flatness_val_ = metric(y, predictions, sample_weight)

    for class_ in [BinBasedSDE, BinBasedTheil, BinBasedCvM]:
        metric = class_(n_bins=n_bins, uniform_features=features, uniform_label=uniform_label, )
        metric.fit(X, y, sample_weight=sample_weight)
        flatness_val_ = metric(y, predictions, sample_weight)



from __future__ import division, print_function, absolute_import
import copy
import numpy
from sklearn.metrics import mean_squared_error, roc_auc_score
from sklearn.base import clone
from hep_ml.commonutils import generate_sample
from hep_ml.losses import LogLossFunction, MSELossFunction, AdaLossFunction
from hep_ml import losses
from hep_ml.gradientboosting import UGradientBoostingClassifier, UGradientBoostingRegressor


def test_gb_with_ada_and_log(n_samples=1000, n_features=10, distance=0.6):
    """
    Testing with two main classification losses.
    Also testing copying
    """
    testX, testY = generate_sample(n_samples, n_features, distance=distance)
    trainX, trainY = generate_sample(n_samples, n_features, distance=distance)
    for loss in [LogLossFunction(), AdaLossFunction()]:
        clf = UGradientBoostingClassifier(loss=loss, min_samples_split=20, max_depth=5, learning_rate=.2,
                                          subsample=0.7, n_estimators=10, train_features=None)
        clf.fit(trainX, trainY)
        assert clf.n_features == n_features
        assert len(clf.feature_importances_) == n_features
        # checking that predict proba works
        for p in clf.staged_predict_proba(testX):
            assert p.shape == (n_samples, 2)
        assert numpy.all(p == clf.predict_proba(testX))
        assert roc_auc_score(testY, p[:, 1]) > 0.8, 'quality is too low'
        # checking clonability
        _ = clone(clf)
        clf_copy = copy.deepcopy(clf)
        assert numpy.all(clf.predict_proba(trainX) == clf_copy.predict_proba(trainX)), 'copied classifier is different'


def test_gradient_boosting(n_samples=1000, distance = 0.6):
    """
    Testing workability of GradientBoosting with different loss function
    """
    # Generating some samples correlated with first variable
    testX, testY = generate_sample(n_samples, 10, distance)
    trainX, trainY = generate_sample(n_samples, 10, distance)
    # We will try to get uniform distribution along this variable
    uniform_features = ['column0']

    loss1 = LogLossFunction()
    loss2 = AdaLossFunction()
    loss3 = losses.CompositeLossFunction()
    loss4 = losses.KnnAdaLossFunction(uniform_features=uniform_features, knn=5, uniform_label=1)
    loss5 = losses.KnnAdaLossFunction(uniform_features=uniform_features, knn=5, uniform_label=[0, 1])
    loss6bin = losses.BinFlatnessLossFunction(uniform_features, fl_coefficient=1., uniform_label=0)
    loss7bin = losses.BinFlatnessLossFunction(uniform_features, fl_coefficient=1., uniform_label=[0, 1])
    loss6knn = losses.KnnFlatnessLossFunction(uniform_features, fl_coefficient=1., uniform_label=1)
    loss7knn = losses.KnnFlatnessLossFunction(uniform_features, fl_coefficient=1., uniform_label=[0, 1])

    for loss in [loss1, loss2, loss3, loss4, loss5, loss6bin, loss7bin, loss6knn, loss7knn]:
        clf = UGradientBoostingClassifier(loss=loss, min_samples_split=20, max_depth=5, learning_rate=0.2,
                                          subsample=0.7, n_estimators=25, train_features=None)
        clf.fit(trainX[:n_samples], trainY[:n_samples])
        result = clf.score(testX, testY)
        assert result >= 0.7, "The quality is too poor: {} with loss: {}".format(result, loss)

    trainX['fake_request'] = numpy.random.randint(0, 4, size=len(trainX))
    testX['fake_request'] = numpy.random.randint(0, 4, size=len(testX))
    for loss in [losses.MSELossFunction(),
                 losses.MAELossFunction(),
                 losses.RankBoostLossFunction(request_column='fake_request')]:
        print(loss)
        clf = UGradientBoostingRegressor(loss=loss, max_depth=3, n_estimators=50, learning_rate=0.01, subsample=0.5,
                                         train_features=list(trainX.columns[1:]))
        clf.fit(trainX, trainY)
        roc_auc = roc_auc_score(testY, clf.predict(testX))
        assert roc_auc >= 0.7, "The quality is too poor: {} with loss: {}".format(roc_auc, loss)


def test_gb_regression(n_samples=1000):
    X, _ = generate_sample(n_samples, 10, distance=0.6)
    y = numpy.tanh(X.sum(axis=1))
    clf = UGradientBoostingRegressor(loss=MSELossFunction())
    clf.fit(X, y)
    y_pred = clf.predict(X)
    zeromse = 0.5 * mean_squared_error(y, y * 0.)
    assert mean_squared_error(y, y_pred) < zeromse, 'something wrong with regression quality'


def test_gb_ranking(n_samples=1000):
    """
    Testing RankingLossFunction
    """
    distance = 0.6
    testX, testY = generate_sample(n_samples, 10, distance)
    trainX, trainY = generate_sample(n_samples, 10, distance)

    rank_variable = 'column1'
    trainX[rank_variable] = numpy.random.randint(0, 3, size=len(trainX))
    testX[rank_variable] = numpy.random.randint(0, 3, size=len(testX))

    rank_loss1 = losses.RankBoostLossFunction(request_column=rank_variable, update_iterations=1)
    rank_loss2 = losses.RankBoostLossFunction(request_column=rank_variable, update_iterations=2)
    rank_loss3 = losses.RankBoostLossFunction(request_column=rank_variable, update_iterations=10)

    for loss in [rank_loss1, rank_loss2, rank_loss3]:
        clf = UGradientBoostingRegressor(loss=loss, min_samples_split=20, max_depth=5, learning_rate=0.2,
                                         subsample=0.7, n_estimators=25, train_features=None) \
            .fit(trainX[:n_samples], trainY[:n_samples])
        result = roc_auc_score(testY, clf.predict(testX))
        assert result >= 0.8, "The quality is too poor: {} with loss: {}".format(result, loss)


def test_constant_fitting(n_samples=1000, n_features=5):
    """
    Testing if initial constant fitted properly
    """
    X, y = generate_sample(n_samples=n_samples, n_features=n_features)
    y = y.astype(float) + 1000.
    for loss in [MSELossFunction(), losses.MAELossFunction()]:
        gb = UGradientBoostingRegressor(loss=loss, n_estimators=10)
        gb.fit(X, y)
        p = gb.predict(X)
        assert mean_squared_error(p, y) < 0.5


def test_weight_misbalance(n_samples=1000, n_features=10, distance=0.6):
    """
    Testing how classifiers work with highly misbalanced (in the terms of weights) datasets.
    """
    testX, testY = generate_sample(n_samples, n_features, distance=distance)
    trainX, trainY = generate_sample(n_samples, n_features, distance=distance)
    trainW = trainY * 10000 + 1
    testW = testY * 10000 + 1
    for loss in [LogLossFunction(), AdaLossFunction(), losses.CompositeLossFunction()]:
        clf = UGradientBoostingClassifier(loss=loss, min_samples_split=20, max_depth=5, learning_rate=.2,
                                          subsample=0.7, n_estimators=10, train_features=None)
        clf.fit(trainX, trainY, sample_weight=trainW)
        p = clf.predict_proba(testX)
        assert roc_auc_score(testY, p[:, 1], sample_weight=testW) > 0.8, 'quality is too low'
